import itertools
import logging
from abc import ABC, abstractmethod
from collections import deque
from typing import List, Tuple, Optional

import numpy as np

from somax.features import MaxVelocity, VerticalDensity
from somax.features.spectral_features import OctaveBands
from somax.runtime.corpus import Corpus
from somax.runtime.corpus_event import CorpusEvent
from somax.runtime.improvisation_memory import ImprovisationMemory
from somax.runtime.parameter import Parametric, Parameter, ParamWithSetter
from somax.runtime.peaks import Peaks
from somax.runtime.transform_handler import TransformHandler
from somax.runtime.transforms import AbstractTransform
from somax.utils.introspective import StringParsed


class AbstractScaleAction(Parametric, StringParsed, ABC):
    def __init__(self):
        super().__init__()
        self.enabled: Parameter = Parameter(True, False, True, "bool", "Enables this ScaleAction.")

    @abstractmethod
    def scale(self, peaks: Peaks, time: float, corresponding_events: List[CorpusEvent],
              corresponding_transforms: List[AbstractTransform], history: ImprovisationMemory = None,
              corpus: Corpus = None, **kwargs) -> Peaks:
        """ """

    @abstractmethod
    def feedback(self, feedback_event: Optional[CorpusEvent], time: float,
                 applied_transform: AbstractTransform) -> None:
        """ """

    @abstractmethod
    def update_transforms(self, transform_handler: TransformHandler):
        """ """

    @abstractmethod
    def clear(self) -> None:
        """ """

    @classmethod
    def default(cls, **_kwargs) -> 'AbstractScaleAction':
        return NoScaleAction()

    @classmethod
    def default_set(cls, **_kwargs) -> Tuple['AbstractScaleAction']:
        return PhaseModulationScaleAction(),

    @classmethod
    def from_string(cls, scale_action: str, **kwargs) -> 'AbstractScaleAction':
        return cls._from_string(scale_action, **kwargs)

    def is_enabled(self):
        return self.enabled.value


class NoScaleAction(AbstractScaleAction):
    def __init__(self):
        super().__init__()

    def scale(self, peaks: Peaks, _time: float, _corresponding_events: List[CorpusEvent],
              _corresponding_transforms: List[AbstractTransform], _history: ImprovisationMemory = None,
              _corpus: Corpus = None, **_kwargs) -> Peaks:
        return peaks

    def feedback(self, feedback_event: Optional[CorpusEvent], time: float,
                 applied_transform: AbstractTransform) -> None:
        pass

    def update_transforms(self, transform_handler: TransformHandler):
        pass

    def clear(self) -> None:
        pass


class PhaseModulationScaleAction(AbstractScaleAction):
    DEFAULT_SELECTIVITY = 3.0

    def __init__(self, selectivity=DEFAULT_SELECTIVITY):
        super().__init__()
        self.logger = logging.getLogger(__name__)
        self.logger.debug("[__init__] Creating PhaseMergeAction with selectivity {}".format(selectivity))
        self._selectivity: Parameter = Parameter(selectivity, None, None, 'float', "Very unclear parameter.")  # TODO
        self._parse_parameters()

    def scale(self, peaks: Peaks, time: float, _corresponding_events: List[CorpusEvent],
              _corresponding_transforms: List[AbstractTransform], _history: ImprovisationMemory = None,
              _corpus: Corpus = None, **_kwargs) -> Peaks:
        peaks.scores *= np.exp(self.selectivity * (np.cos(2 * np.pi * (time - peaks.times)) - 1))
        return peaks

    def feedback(self, _feedback_event: Optional[CorpusEvent], _time: float,
                 _applied_transform: AbstractTransform) -> None:
        pass

    def update_transforms(self, transform_handler: TransformHandler):
        pass

    def clear(self) -> None:
        pass

    @property
    def selectivity(self):
        return self._selectivity.value

    @selectivity.setter
    def selectivity(self, value):
        self._selectivity.value = value


class NextStateScaleAction(AbstractScaleAction):
    DEFAULT_FACTOR = 1.5

    def __init__(self, factor: float = DEFAULT_FACTOR):
        super().__init__()
        self.logger = logging.getLogger(__name__)
        self.logger.debug(f"[__init__] Creating {type(self).__name__} with factor {factor}.")
        self._factor: Parameter = Parameter(factor, 0.0, None, 'float',
                                            "Scaling factor for peaks close to previous output.")
        self._previous_output_index: Optional[int] = None

    def scale(self, peaks: Peaks, time: float, corresponding_events: List[CorpusEvent],
              _corresponding_transforms: List[AbstractTransform], history: ImprovisationMemory = None,
              corpus: Corpus = None, **_kwargs) -> Peaks:
        if self._previous_output_index is None:
            return peaks
        else:
            event_indices: np.ndarray = np.array([e.state_index for e in corresponding_events], dtype=int)
            is_matching: np.ndarray = event_indices == self._previous_output_index + 1
            peaks.scale(self.factor, is_matching)
            return peaks

    def feedback(self, feedback_event: Optional[CorpusEvent], _time: float,
                 _applied_transform: AbstractTransform) -> None:
        if feedback_event is not None:
            self._previous_output_index = feedback_event.state_index

    def update_transforms(self, transform_handler: TransformHandler):
        pass

    def clear(self) -> None:
        self._previous_output_index = None

    @property
    def factor(self):
        return self._factor.value


class AutoJumpScaleAction(AbstractScaleAction):
    DEFAULT_ACTIVATION_THRESHOLD = 2
    DEFAULT_JUMP_THRESHOLD = 8

    def __init__(self, activation_threshold: int = DEFAULT_ACTIVATION_THRESHOLD,
                 jump_threshold: int = DEFAULT_JUMP_THRESHOLD):
        super().__init__()
        self._activation_threshold: Parameter = Parameter(activation_threshold, 1, None, "int", "TODO")
        self._jump_threshold: Parameter = Parameter(jump_threshold, 1, None, "int", "TODO")

    def scale(self, peaks: Peaks, time: float, corresponding_events: List[CorpusEvent],
              corresponding_transforms: List[AbstractTransform], history: ImprovisationMemory = None,
              corpus: Corpus = None, **kwargs) -> Peaks:
        event_indices: List[int] = [e[0].state_index for e in history.get_n_latest(self.jump_threshold + 1)]
        if not event_indices:
            return peaks
        previous_index: int = event_indices[0]
        num_consecutive_events: int = len(list(itertools.takewhile(lambda a: a == -1, np.diff(event_indices))))
        if num_consecutive_events <= self.activation_threshold:
            factor: float = 1.0
        elif num_consecutive_events >= self.jump_threshold:
            factor: float = 0.0
        else:
            factor: float = 0.5 ** (num_consecutive_events - self.activation_threshold)
        event_indices: np.ndarray = np.array([e.state_index for e in corresponding_events], dtype=int)
        is_matching: np.ndarray = event_indices == previous_index + 1
        peaks.scale(factor, is_matching)
        return peaks

    def feedback(self, feedback_event: Optional[CorpusEvent], time: float,
                 applied_transform: AbstractTransform) -> None:
        pass

    def update_transforms(self, transform_handler: TransformHandler):
        pass

    def clear(self) -> None:
        pass

    @property
    def activation_threshold(self):
        return self._activation_threshold.value

    @property
    def jump_threshold(self):
        return self._jump_threshold.value


class TempoConsistencyScaleAction(AbstractScaleAction):
    DEFAULT_NUM_EVENTS: int = 5
    DEFAULT_SIGMA: float = 5.0

    def __init__(self, history_len: int = DEFAULT_NUM_EVENTS, sigma: float = DEFAULT_SIGMA):
        super().__init__()
        self._history_len: Parameter = Parameter(history_len, 1, None, "int", "TODO")
        self._sigma: Parameter = Parameter(sigma, None, None, 'float', "Standard deviation of gaussian")

    def scale(self, peaks: Peaks, time: float, corresponding_events: List[CorpusEvent],
              corresponding_transforms: List[AbstractTransform], history: ImprovisationMemory = None,
              corpus: Corpus = None, **kwargs) -> Peaks:
        previous_tempi: np.ndarray = np.array([e[0].tempo for e in history.get_n_latest(self.history_len)])
        if previous_tempi.size == 0:
            return peaks
        mu: float = float(np.mean(previous_tempi))
        peaks_tempi: np.ndarray = np.array([e.tempo for e in corresponding_events])
        peaks.scores *= np.exp(-((peaks_tempi - mu) ** 2 / (2 * self.sigma ** 2)))
        return peaks

    def feedback(self, feedback_event: Optional[CorpusEvent], time: float,
                 applied_transform: AbstractTransform) -> None:
        pass

    def update_transforms(self, transform_handler: TransformHandler):
        pass

    def clear(self) -> None:
        pass

    @property
    def history_len(self):
        return self._history_len.value

    @property
    def sigma(self):
        return self._sigma.value


class StaticTabooScaleAction(AbstractScaleAction):
    DEFAULT_TABOO_LENGTH = 10

    def __init__(self, taboo_length: int = DEFAULT_TABOO_LENGTH):
        super().__init__()
        self.logger = logging.getLogger(__name__)
        self._taboo_length: Parameter = ParamWithSetter(taboo_length, 1, None, 'int',
                                                        "Number of events to pass before event can be played again",
                                                        self._set_taboo_length)

        self._taboo_indices: deque[int] = deque([], self.taboo_length)

    def scale(self, peaks: Peaks, time: float, corresponding_events: List[CorpusEvent],
              corresponding_transforms: List[AbstractTransform], history: ImprovisationMemory = None,
              corpus: Corpus = None, **kwargs) -> Peaks:
        event_indices: np.ndarray = np.array([e.state_index for e in corresponding_events], dtype=int)
        matching_indices: np.ndarray = np.zeros(len(corresponding_events), dtype=bool)
        for taboo_index in self._taboo_indices:
            matching_indices += event_indices == taboo_index
        peaks.scale(0, matching_indices)
        return peaks

    def feedback(self, feedback_event: Optional[CorpusEvent], _time: float,
                 _applied_transform: AbstractTransform) -> None:
        if feedback_event is not None:
            self._taboo_indices.append(feedback_event.state_index)
        else:
            self._taboo_indices.append(-1)

    def update_transforms(self, transform_handler: TransformHandler):
        pass

    def clear(self) -> None:
        self._taboo_indices: deque[int] = deque([], self.taboo_length)

    @property
    def taboo_length(self) -> int:
        return int(self._taboo_length.value)

    def _set_taboo_length(self, n: int):
        self._taboo_length.value = n
        self.clear()


class BinaryTransformContinuityScaleAction(AbstractScaleAction):
    DEFAULT_FACTOR = 0.5

    def __init__(self, factor: float = DEFAULT_FACTOR):
        super().__init__()
        self.logger = logging.getLogger(__name__)
        self._factor: Parameter = Parameter(factor, 0.0, None, 'float',
                                            "Scaling factor for peaks not matching previous transform")
        self._previous_transform: Optional[AbstractTransform] = None
        # This will always be set immediately after adding transform to player and the
        # Optional case should never have to be handled
        self._transform_handler: Optional[TransformHandler] = None

    def scale(self, peaks: Peaks, time: float, corresponding_events: List[CorpusEvent],
              corresponding_transforms: List[AbstractTransform], history: ImprovisationMemory = None,
              corpus: Corpus = None, **kwargs) -> Peaks:
        if self._previous_transform is None or self._transform_handler:
            return peaks
        else:
            peak_transform_ids: np.ndarray = np.array(peaks.transform_ids)
            previous_transform_id: int = self._transform_handler.get_id(self._previous_transform)
            not_matching: np.ndarray = peak_transform_ids != previous_transform_id
            peaks.scale(self.factor, not_matching)
            return peaks

    def feedback(self, _feedback_event: Optional[CorpusEvent], _time: float,
                 applied_transform: AbstractTransform) -> None:
        if _feedback_event is not None:
            self._previous_transform = applied_transform

    def update_transforms(self, transform_handler: TransformHandler):
        self._transform_handler = transform_handler

    def clear(self) -> None:
        self._previous_transform = None

    @property
    def factor(self):
        return self._factor.value


class AbstractGaussianScale(AbstractScaleAction, ABC):
    def __init__(self, mu: float = 0.0, sigma: float = 1.0):
        super().__init__()
        self._mu: Parameter = Parameter(mu, None, None, 'float', "Mean value of gaussian.")
        self._sigma: Parameter = Parameter(sigma, None, None, 'float', "Standard deviation of gaussian")

    def _scale(self, peaks: Peaks, corresponding_features: np.ndarray) -> Peaks:
        peaks.scores *= np.exp(-((corresponding_features - self.mu) ** 2 / (2 * self.sigma ** 2)))
        return peaks

    @property
    def mu(self):
        return self._mu.value

    @property
    def sigma(self):
        return self._sigma.value


class MaxVelocityScaleAction(AbstractGaussianScale):
    DEFAULT_VELOCITY = 0.5

    def __init__(self, mu: float = DEFAULT_VELOCITY):
        super().__init__(mu=mu)

    def scale(self, peaks: Peaks, time: float, corresponding_events: List[CorpusEvent],
              _corresponding_transforms: List[AbstractTransform], _history: ImprovisationMemory = None,
              _corpus: Corpus = None, **_kwargs) -> Peaks:
        velocities: np.ndarray = np.array([event.get_feature(MaxVelocity).value() for event in corresponding_events])
        return self._scale(peaks, velocities)

    def feedback(self, feedback_event: Optional[CorpusEvent], time: float,
                 applied_transform: AbstractTransform) -> None:
        pass

    def update_transforms(self, transform_handler: TransformHandler):
        pass

    def clear(self) -> None:
        pass


class VerticalDensityScaleAction(AbstractGaussianScale):
    DEFAULT_VERTICAL_DENSITY = 4

    def __init__(self, mu: float = DEFAULT_VERTICAL_DENSITY):
        super().__init__(mu=mu)

    def scale(self, peaks: Peaks, time: float, corresponding_events: List[CorpusEvent],
              _corresponding_transforms: List[AbstractTransform], _history: ImprovisationMemory = None,
              _corpus: Corpus = None, **_kwargs) -> Peaks:
        densities: np.ndarray = np.array([event.get_feature(VerticalDensity).value() for event in corresponding_events])
        return self._scale(peaks, densities)

    def feedback(self, feedback_event: Optional[CorpusEvent], time: float,
                 applied_transform: AbstractTransform) -> None:
        pass

    def update_transforms(self, transform_handler: TransformHandler):
        pass

    def clear(self) -> None:
        pass


class DurationScaleAction(AbstractGaussianScale):
    DEFAULT_DURATION = 1.0

    def __init__(self, mu: float = DEFAULT_DURATION):
        super().__init__(mu=mu)

    def scale(self, peaks: Peaks, time: float, corresponding_events: List[CorpusEvent],
              corresponding_transforms: List[AbstractTransform], history: ImprovisationMemory = None,
              corpus: Corpus = None, **kwargs) -> Peaks:
        durations: np.ndarray = np.array([event.duration for event in corresponding_events])
        return self._scale(peaks, durations)

    def feedback(self, feedback_event: Optional[CorpusEvent], time: float,
                 applied_transform: AbstractTransform) -> None:
        pass

    def update_transforms(self, transform_handler: TransformHandler):
        pass

    def clear(self) -> None:
        pass


# TODO: Update so that it takes transforms into account
class OctaveBandsScaleAction(AbstractScaleAction):
    DEFAULT_BAND_DISTRIBUTION = np.ones(OctaveBands.NUM_BANDS, dtype=float)

    def __init__(self):
        super().__init__()
        self._band_distribution: Parameter = ParamWithSetter(OctaveBandsScaleAction.DEFAULT_BAND_DISTRIBUTION, None,
                                                             None, "list[11]", "TODO", self._set_band_distribution)

    def scale(self, peaks: Peaks, time: float, corresponding_events: List[CorpusEvent],
              corresponding_transforms: List[AbstractTransform], history: ImprovisationMemory = None,
              corpus: Corpus = None, **kwargs) -> Peaks:
        events_band_distribution: np.ndarray = np.array([event.get_feature(OctaveBands).value()
                                                         for event in corresponding_events])
        factor: np.ndarray = 1 / np.sqrt(np.sum(np.power(events_band_distribution - self.band_distribution, 2), axis=1))
        print(np.min(factor), np.max(factor), factor.shape)  # TODO: THIS SHOULD BE HERE UNTIL PROPERLY DEBUGGED
        peaks.scale(factor)
        return peaks

    def feedback(self, feedback_event: Optional[CorpusEvent], time: float,
                 applied_transform: AbstractTransform) -> None:
        pass

    def update_transforms(self, transform_handler: TransformHandler):
        pass  # TODO: Handle transforms

    def clear(self) -> None:
        pass

    @property
    def band_distribution(self):
        return self._band_distribution.value

    def _set_band_distribution(self, band_distribution: List[float]):
        band_distribution: np.ndarray = np.array(band_distribution, dtype=float)
        max_val: float = np.max(band_distribution)
        if max_val > 0:
            band_distribution /= max_val
        self._band_distribution.value = band_distribution


class RegionMaskScaleAction(AbstractScaleAction):
    def __init__(self):
        super().__init__()
        self._low_thresh: Parameter = Parameter(0, 0, 1.0, "float", "Fraction [0,1] marking start of region")
        self._high_thresh: Parameter = Parameter(1.0, 0, 1.0, "float", "Fraction [0,1] marking end of region")

    def scale(self, peaks: Peaks, _time: float, corresponding_events: List[CorpusEvent],
              corresponding_transforms: List[AbstractTransform], _history: ImprovisationMemory = None,
              corpus: Corpus = None, **_kwargs) -> Peaks:
        # TODO: This could be optimized and stored if ScaleAction had direct access to Corpus
        low_index: int = int(self._low_thresh.value * corpus.length())
        high_index: int = int(self._high_thresh.value * corpus.length())
        corresponding_indices: np.ndarray = np.array([e.state_index for e in corresponding_events], dtype=int)
        mask: np.ndarray = ((low_index <= corresponding_indices) & (corresponding_indices <= high_index)).astype(int)
        peaks.scale(mask)
        return peaks

    def feedback(self, feedback_event: Optional[CorpusEvent], time: float,
                 applied_transform: AbstractTransform) -> None:
        pass

    def update_transforms(self, transform_handler: TransformHandler):
        pass

    def clear(self) -> None:
        pass



from abc import ABC
from typing import Callable, Any, Dict, Optional, List

from somax.runtime.atom import Atom
from somax.runtime.corpus_event import CorpusEvent
from somax.runtime.influence import AbstractInfluence
from somax.runtime.transforms import AbstractTransform


class ScheduledEvent(ABC):
    def __init__(self, trigger_time: float):
        self.trigger_time = trigger_time


class TempoEvent(ScheduledEvent):
    def __init__(self, trigger_time: float, tempo: float):
        super(TempoEvent, self).__init__(trigger_time)
        self.tempo = tempo


class MidiEvent(ScheduledEvent):
    def __init__(self, trigger_time: float, note: int, velocity: int, channel: int, state: Optional[int] = None,
                 applied_transform: Optional[AbstractTransform] = None):
        super(MidiEvent, self).__init__(trigger_time)
        self.note: int = note
        self.velocity: int = velocity
        self.channel: int = channel
        self.state: Optional[int] = state
        self.applied_transform: Optional[AbstractTransform] = applied_transform

    def __repr__(self):
        return f"MidiEvent(trigger_time={self.trigger_time},note={self.note},velocity={self.velocity}," \
               f"channel={self.channel},state={self.state})"


class NewStateEvent(ScheduledEvent):
    """ Class used to send state information as well as a `chord` message to treat all note onsets in the slice as
        simultaneous and thereby bypass quickthresh latency when re-influencing other player """
    def __init__(self, trigger_time: float, corpus_event: CorpusEvent, applied_transform: AbstractTransform):
        super().__init__(trigger_time=trigger_time)
        self.corpus_event: CorpusEvent = corpus_event
        self.applied_transform: AbstractTransform = applied_transform


class AudioEvent(ScheduledEvent):
    def __init__(self, trigger_time: float, onset: float, duration: float, state: int, tempo: float):
        super(AudioEvent, self).__init__(trigger_time)
        self.onset: float = onset
        self.duration: float = duration
        self.state: int = state
        self.tempo: float = tempo


class ScheduledCorpusEvent(ScheduledEvent):
    def __init__(self, trigger_time: float, corpus_event: CorpusEvent):
        super().__init__(trigger_time)
        self.corpus_event: CorpusEvent = corpus_event


class ScheduledInfluenceEvent(ScheduledEvent):
    def __init__(self, trigger_time: float, path: List[str], influence: AbstractInfluence):
        super().__init__(trigger_time)
        self.path = path
        self.influence = influence
        self.num_generated_peaks: Optional[Dict[Atom, int]] = None  # added to event after it's been processed


class AbstractTriggerEvent(ScheduledEvent, ABC):
    def __init__(self, trigger_time: float, target_time: float):
        super(AbstractTriggerEvent, self).__init__(trigger_time)
        self.target_time: float = target_time


class AutomaticTriggerEvent(AbstractTriggerEvent):
    def __init__(self, trigger_time: float, target_time: float):
        super().__init__(trigger_time, target_time)

    def __repr__(self):
        return f"AutomaticTriggerEvent(trigger_time={self.trigger_time},target_time={self.target_time}"


class ManualTriggerEvent(AbstractTriggerEvent):
    def __init__(self, trigger_time: float):
        super(ManualTriggerEvent, self).__init__(trigger_time, trigger_time)


class OscEvent(ScheduledEvent):
    pass  # TODO


class CallableEvent(ScheduledEvent):
    def __init__(self, trigger_time: float, func: Callable, **kwargs):
        super().__init__(trigger_time)
        self.func: Callable = func
        self.kwargs: Dict[str, Any] = kwargs



import logging
from enum import Enum
from typing import Optional, List, Iterator, Any, Iterable, Tuple, Dict

import mido
import numpy as np
import pandas as pd
from mido import MidiTrack, Message, MetaMessage, MidiFile

from .matrix_keys import MatrixKeys as Keys


class BarNumberAnnotation(Enum):
    NONE = "none"
    JUMPS = "jumps"
    ALL = "all"

    @classmethod
    def default(cls) -> 'BarNumberAnnotation':
        return cls.NONE

    @classmethod
    def from_string(cls, bar_num_annotation: str) -> 'BarNumberAnnotation':
        try:
            return BarNumberAnnotation(bar_num_annotation.lower())
        except ValueError:
            logging.getLogger(__name__).warning(f"No class named '{bar_num_annotation} exists for the "
                                                f"{cls.__name__} module. Using default.")
            return cls.default()


class _MidiNote:
    """ Temporary class used only while constructing the NoteMatrix"""

    def __init__(self, note: int, vel: int, ch: int, onset_tick: int, onset_time: float, tempo: float, bar: float,
                 track: str, end_tick: Optional[int] = None, end_time: Optional[float] = None):
        self.note: int = note
        self.vel: int = vel
        self.ch: int = ch
        self.onset_tick: int = onset_tick
        self.onset_time: float = onset_time
        self.tempo: float = tempo
        self.bar: float = bar
        self.track: str = track
        self.end_tick: Optional[int] = end_tick
        self.end_time: Optional[float] = end_time

    def matches(self, note: int, ch: int, track: str) -> bool:
        return self.note == note and self.ch == ch and self.track == track


class MidiParser:
    _NUM_COLS = len(Keys)
    _DEFAULT_TEMPO_BPM = 120

    @staticmethod
    def read_midi(midi_file_path: str) -> pd.DataFrame:
        midi_file: mido.MidiFile = mido.MidiFile(midi_file_path)
        notes: List[_MidiNote] = MidiParser._parse_notes(midi_file)
        note_matrix: pd.DataFrame = MidiParser._to_pandas(notes, midi_file.ticks_per_beat)
        return note_matrix

    @staticmethod
    def export_midi(note_matrix: pd.DataFrame, filepath: str, title: str, initial_time_signature: Tuple[int, int],
                    ticks_per_beat: int,
                    annotations: BarNumberAnnotation):
        midi_notes, duration = MidiParser._from_pandas(note_matrix, ticks_per_beat)  # type: List[_MidiNote], int
        midi_tracks: List[MidiTrack] = MidiParser._serialize_notes(midi_notes, duration, title,
                                                                   initial_time_signature=initial_time_signature,
                                                                   annotations=annotations)
        midi_file: MidiFile = MidiFile(ticks_per_beat=ticks_per_beat)
        for track in midi_tracks:
            midi_file.tracks.append(track)

        midi_file.save(filepath)
        return midi_file

    @staticmethod
    def monophonic_from_text(note_numbers: np.ndarray, velocities: np.ndarray, channels: np.ndarray,
                             durations_tick: np.ndarray, tempo: float):
        """ Notes: durations_tick are float point ticks, i.e. 1.0 is one quarter note """
        notes: List[_MidiNote] = []
        ppq: int = 480
        onset_tick: int = 0
        onset_time: float = 0.0
        for nn, vel, ch, dur in zip(note_numbers, velocities, channels, durations_tick):
            quantized_duration_tick: int = round(dur * ppq)
            duration_sec: float = quantized_duration_tick / ppq * 60 / tempo
            notes.append(_MidiNote(nn, vel, ch, onset_tick, onset_time, tempo, 0, "",
                                   onset_tick + quantized_duration_tick, onset_time + duration_sec))
            onset_tick += quantized_duration_tick
            onset_time += duration_sec
        return MidiParser._to_pandas(notes, ppq)

    @staticmethod
    def _parse_notes(midi_file: mido.MidiFile) -> List[_MidiNote]:
        ticks_per_beat: int = midi_file.ticks_per_beat
        tempo_bpm: float = MidiParser._DEFAULT_TEMPO_BPM
        held_notes: List[_MidiNote] = []
        completed_notes: List[_MidiNote] = []
        current_tick: int = 0
        current_time: float = 0.0
        current_bar: float = 1.0
        current_bar_factor: int = 1

        for msg, track in MidiParser.__merge_tracks(midi_file.tracks):
            current_tick += msg.time
            current_time += mido.tick2second(msg.time, ticks_per_beat, mido.bpm2tempo(tempo_bpm))
            current_bar += msg.time / ticks_per_beat * current_bar_factor
            track_name: str = track.name if track else ""
            if msg.type == 'set_tempo':
                tempo_bpm = mido.tempo2bpm(msg.tempo)
            elif msg.type == 'time_signature':
                current_bar_factor = msg.denominator / (4 * msg.numerator)
            elif msg.type == 'note_on' or msg.type == 'note_off':
                # if the note is already held, complete the note regardless of whether the input is a note_on or
                # note_off message. This is to avoid duplicate notes in a slice if re-triggered before its note_off
                # See test case "keithjarrett_kolnconcert_Right.mid" bars 311 to 330 for an example of this.
                completed: List[_MidiNote] = [note for note in held_notes
                                              if note.matches(msg.note, msg.channel, track_name)]
                for note in completed:
                    note.end_tick = current_tick
                    note.end_time = current_time
                completed_notes.extend(completed)
                held_notes = [note for note in held_notes if not note.matches(msg.note, msg.channel, track_name)]
                if msg.type == 'note_on' and msg.velocity > 0:
                    held_notes.append(_MidiNote(msg.note, msg.velocity, msg.channel, current_tick,
                                                current_time, tempo_bpm, current_bar, track_name))
                # elif msg.type == 'note_off' or (msg.type == 'note_on' and msg.velocity == 0):

        for note in held_notes:
            # If note ons still exist at end of midi file, generate note offs for all of them
            note.end_tick = current_tick
            note.end_time = current_time
            completed_notes.append(note)

        return completed_notes

    @staticmethod
    def _serialize_notes(notes: List[_MidiNote], duration: int, title: str, initial_time_signature: Tuple[int, int],
                         annotations: BarNumberAnnotation, **kwargs) -> List[MidiTrack]:
        meta_export_dict: str = "meta"
        lyrics_dict: str = "lyrics"
        messages: Dict[str, List[Any]] = {
            meta_export_dict: [
                MetaMessage(type='track_name', name=title, time=0),
                MetaMessage(type='time_signature',
                            numerator=initial_time_signature[0],
                            denominator=initial_time_signature[1],
                            clocks_per_click=12,
                            notated_32nd_notes_per_beat=8,
                            time=0),
            ],
            lyrics_dict: []}

        # Append notes in tick time since start (rather than delta time)
        for note in notes:
            if note.track not in messages:
                messages[note.track] = [MetaMessage(type='track_name', name=note.track)]
            # Note: All `time` fields are appended in absolute time (ticks since start of track, not delta time)
            messages[note.track].append(Message('note_on', note=note.note, velocity=note.vel, channel=note.ch,
                                                time=note.onset_tick))
            messages[note.track].append(Message('note_on', note=note.note, velocity=0, channel=note.ch,
                                                time=note.end_tick))
            messages[meta_export_dict].append(MetaMessage(type='set_tempo', tempo=mido.bpm2tempo(note.tempo),
                                                          time=note.onset_tick))
            messages[lyrics_dict].append(
                MetaMessage(type='lyrics', text=str(int(note.bar + 0.01)), time=note.onset_tick))
        if annotations == BarNumberAnnotation.NONE:
            del messages[lyrics_dict]
        else:
            messages[lyrics_dict].sort(key=lambda msg: msg.time)
            filtered_bar_numbers: List[MetaMessage] = []
            if annotations == BarNumberAnnotation.JUMPS:
                prev_bar_number: int = -np.inf
                for lyrics_meta in messages[lyrics_dict]:
                    cur_bar_number: int = int(lyrics_meta.text)
                    if np.abs(cur_bar_number - prev_bar_number) > 2:
                        filtered_bar_numbers.append(lyrics_meta)
                    prev_bar_number = cur_bar_number

            else:  # annotations == BarNumberAnnotation.ALL:
                last_onset: float = -np.inf
                for lyrics_meta in messages[lyrics_dict]:
                    if lyrics_meta.time - last_onset > 0.0001:
                        filtered_bar_numbers.append(lyrics_meta)
                    last_onset = lyrics_meta.time
            messages[lyrics_dict] = filtered_bar_numbers

        midi_tracks: List[MidiTrack] = []
        for track in messages.keys():
            messages[track].append(MetaMessage(type='end_of_track', time=duration))
            messages[track].sort(key=lambda msg: msg.time)
            messages[track] = MidiParser.__to_reltime(messages[track])
            midi_tracks.append(MidiTrack(messages[track]))

        return midi_tracks

    @staticmethod
    def __merge_tracks(tracks: List[MidiTrack]) -> Iterator[Tuple[Any, Optional[MidiTrack]]]:
        """This function was adapted directly from mido to handle files with uncaught 'type' meta message errors
            as well as append the original track to the sorted list
        """
        messages_and_tracks: List[Tuple[Any, MidiTrack]] = []
        for track in tracks:
            messages_and_tracks.extend([(m, track) for m in MidiParser.__to_abstime(track)])
        messages_and_tracks.sort(key=lambda msg_and_track: msg_and_track[0].time)

        messages, tracks = [list(e) for e in zip(*messages_and_tracks)]
        return MidiParser.__fix_end_of_track(zip(MidiParser.__to_reltime(messages), tracks))

    @staticmethod
    def __fix_end_of_track(messages: Iterable[Tuple[any, MidiTrack]]) -> Iterator[Tuple[Any, Optional[MidiTrack]]]:
        """Remove all end_of_track messages and add one at the end.

        This function was adapted directly from mido to handle passing of track info when merging"""
        # Accumulated delta time from removed end of track messages.
        # This is added to the next message.
        accum = 0

        for msg, track in messages:
            if msg.type == 'end_of_track':
                accum += msg.time
            else:
                if accum:
                    delta = accum + msg.time
                    yield msg.copy(time=delta), track
                    accum = 0
                else:
                    yield msg, track

        yield MetaMessage('end_of_track', time=accum), None

    @staticmethod
    def __to_abstime(messages: MidiTrack) -> Iterable[Any]:
        """Convert messages to absolute time.

        This function was adapted directly from mido to handle files with uncaught 'type' meta message errors"""
        now = 0
        for msg in messages:
            now += msg.time
            try:
                yield msg.copy(time=now)
            except TypeError as e:
                logging.getLogger(__name__).warning(f"Found invalid midi format: {repr(e)}. Error was ignored.")

    @staticmethod
    def __to_reltime(messages: Iterable[Any]) -> Iterable[Any]:
        """Convert messages to relative time.
            This function was adapted directly from mido to handle passing of track info when merging"""
        now = 0
        for msg in messages:
            delta = msg.time - now
            yield msg.copy(time=delta)
            now = msg.time

    @staticmethod
    def _to_pandas(notes: List[_MidiNote], ticks_per_beat: int) -> pd.DataFrame:
        num_notes: int = len(notes)
        note_matrix: np.ndarray = np.zeros((num_notes, MidiParser._NUM_COLS), dtype=object)

        for i, note in enumerate(notes):
            note_number = note.note
            velocity = note.vel
            channel = note.ch + 1
            relative_duration: float = (note.end_tick - note.onset_tick) / ticks_per_beat
            absolute_duration: float = (note.end_time - note.onset_time) * 1000.0
            relative_onset: float = note.onset_tick / ticks_per_beat
            absolute_onset: float = note.onset_time * 1000.0
            tempo_bpm: float = note.tempo
            bar_number: float = note.bar
            track_name: str = note.track
            note_matrix[i] = [note_number, velocity, channel, relative_onset, absolute_onset, relative_duration,
                              absolute_duration, tempo_bpm, bar_number, track_name]

        note_matrix = note_matrix[note_matrix[:, Keys.REL_ONSET.value].argsort()]

        return pd.DataFrame(note_matrix, columns=[key for key in Keys])

    @staticmethod
    def _from_pandas(notes_matrix: pd.DataFrame, ticks_per_beat: int) -> Tuple[List[_MidiNote], float]:
        duration: int = int(np.max(notes_matrix[Keys.REL_ONSET] + notes_matrix[Keys.REL_DURATION]) * ticks_per_beat)
        notes: List[_MidiNote] = []

        # TODO: [OPTIMIZATION] Use list comprehension and call code below as external function for optimization
        for _, note in notes_matrix.iterrows():
            note_number: int = int(note[Keys.PITCH])
            velocity: int = int(note[Keys.VELOCITY])
            channel: int = int(note[Keys.CHANNEL] - 1)
            onset_tick: int = int(note[Keys.REL_ONSET] * ticks_per_beat)
            onset_time: float = note[Keys.ABS_ONSET] * 0.001
            tempo: float = note[Keys.TEMPO]
            bar: float = note[Keys.BAR_NUMBER]
            track: str = note[Keys.TRACK_NAME]
            end_tick: int = int(note[Keys.REL_DURATION] * ticks_per_beat) + onset_tick
            end_time: float = (note[Keys.ABS_DURATION] * 0.001) + onset_time
            notes.append(_MidiNote(note_number, velocity, channel, onset_tick, onset_time, tempo, bar,
                                   track, end_tick, end_time))
        return notes, duration



import logging
from abc import ABC, abstractmethod
from typing import Dict, Type, Any

import numpy as np

from somax.utils.introspective import StringParsed


class AbstractFilter(StringParsed, ABC):
    DEFAULT = "default"
    _DEFAULT_FILTER = "leakyintegrator"

    def __init__(self, decay_length_ms: float = 0.0, **_kwargs):
        """ decay_length_ms is the time until the envelope approaches 0 after an activation
            (required to calculate filter envelope length) """
        self.decay_length_ms: float = decay_length_ms
        self.logger = logging.getLogger(__name__)

    @abstractmethod
    def filter_midi(self, vec: np.ndarray, **kwargs) -> np.ndarray:
        """ vec: 1d or 2d vector that will be filtered """

    @abstractmethod
    def filter_audio(self, **kwargs):
        """ TODO """

    @property
    @abstractmethod
    def build_parameters(self) -> Dict[str, Any]:
        """ TODO """

    @classmethod
    def from_string(cls, class_name: str, **kwargs) -> 'AbstractFilter':
        return cls._from_string(class_name, **kwargs)

    @classmethod
    def default(cls, **kwargs) -> 'AbstractFilter':
        return LeakyIntegrator(**kwargs)

    @staticmethod
    def parse(filter_class: str, **kwargs) -> 'AbstractFilter':
        classes: Dict[str, Type[AbstractFilter]] = AbstractFilter._classes()
        if filter_class == AbstractFilter.DEFAULT:
            return classes[AbstractFilter._DEFAULT_FILTER]()
        try:
            return classes[filter_class](**kwargs)
        except KeyError:
            logging.getLogger(__name__).warning(f"Could not find Filter class {filter_class}. "
                                                f"Using default Filter ({AbstractFilter._DEFAULT_FILTER}) instead.")
            return classes[AbstractFilter._DEFAULT_FILTER]()


class NoFilter(AbstractFilter):
    def __init__(self):
        super().__init__(decay_length_ms=0.0)

    def filter_midi(self, vec: np.ndarray, **_kwargs) -> np.ndarray:
        return vec

    def filter_audio(self, **kwargs):
        raise NotImplementedError("Audio is not supported yet")

    @property
    def build_parameters(self) -> Dict[str, Any]:
        return {"class": self.__class__.__name__}


class LeakyIntegrator(AbstractFilter):
    """Class for filtering 1d activation (binary) vectors with different coefficient for Attack and Release.
        Should be given a vector of ones (activation only), will concatenate release tail duration to vector, i.e.
        f: R^[1xM] -> R^[1x(M+K)] where K denotes activation tail
        # TODO: PROPER DOCSTRING"""

    def __init__(self, filter_amplitude_onset: float = 0.5, filter_amplitude_maximum: float = 1.0,
                 filter_amplitude_threshold: float = 0.05, filter_tau_up: float = 400, filter_tau_down: float = 1000,
                 filter_step_size_ms: float = 20.0, **_kwargs):
        super().__init__(-filter_tau_down * np.log(filter_amplitude_threshold / filter_amplitude_maximum))
        # TODO: Change tau values to actual ms values
        self.amplitude_onset: float = filter_amplitude_onset
        self.amplitude_maximum: float = filter_amplitude_maximum
        self.tau_up: float = filter_tau_up
        self.tau_down: float = filter_tau_down
        self.step_size_ms: float = filter_step_size_ms

    def filter_midi(self, vec: np.ndarray, **kwargs) -> np.ndarray:
        # TODO: Self.amplitude maximum at end doesn't seem correct.
        amplitude_up: np.ndarray = (self.amplitude_onset - self.amplitude_maximum) \
                                   * np.exp(-(np.arange(vec.size) * self.step_size_ms) / self.tau_up) \
                                   + self.amplitude_maximum
        amplitude_down = amplitude_up[-1] * np.exp(-(np.arange(0, self.decay_length_ms, self.step_size_ms)
                                                     / self.tau_down))
        return np.concatenate((amplitude_up, amplitude_down))

    def filter_audio(self, **kwargs):
        raise NotImplementedError("Audio is not supported yet")

    @property
    def build_parameters(self) -> Dict[str, Any]:
        return {"filter_class": self.__class__.__name__,
                "filter_amplitude_onset": self.amplitude_onset,
                "filter_amplitude_maximum": self.amplitude_maximum,
                "filter_tau_up": self.tau_up,
                "filter_tau_down": self.tau_down,
                "filter_step_size_ms": self.step_size_ms}



from .chroma_classifiers import SomChromaClassifier, AbsoluteGmmClassifier, RelativeGmmClassifier
from .pitch_classifiers import TopNoteClassifier, PitchClassClassifier, VirtualFundamentalClassifier, \
    BassNoteClassifier, BassNoteMod12Classifier



import os

PROJECT_ROOT = os.path.abspath(os.path.dirname(__file__))
CORPUS_FOLDER = os.path.join(PROJECT_ROOT, "..", "corpus")

PATH_SEPARATOR = "::"

IMPORT_MATPLOTLIB = False

DEBUG = True

LOGGING_CONFIG = {
    # TODO
}





from enum import Enum


class MatrixKeys(Enum):
    PITCH = 0
    VELOCITY = 1
    CHANNEL = 2
    REL_ONSET = 3
    ABS_ONSET = 4
    REL_DURATION = 5
    ABS_DURATION = 6
    TEMPO = 7
    BAR_NUMBER = 8
    TRACK_NAME = 9



from abc import ABC
from enum import Enum

from somax.scheduler.transport import Time


class PlayControl(Enum):
    START = 0
    PAUSE = 1
    STOP = 2
    CLEAR = 3
    TERMINATE = 4


class ProcessMessage(ABC):
    pass


class TempoMessage(ProcessMessage):
    """ Message from Agent to Server """
    def __init__(self, tempo: float):
        self.tempo: float = tempo


class TimeMessage(ProcessMessage):
    """ Message from Server to Agent """
    def __init__(self, time: Time):
        self.time: Time = time


class TempoMasterMessage(ProcessMessage):
    """ Message from Server to Agent """
    def __init__(self, is_master: bool):
        self.is_master = is_master


class ControlMessage(ProcessMessage):
    """ Message from Server to Agent """
    def __init__(self, msg: PlayControl):
        self.msg = msg



import logging
from typing import Optional, List, Dict, Type, Any

import pandas as pd

from somax.corpus_builder.matrix_keys import MatrixKeys as Keys
from somax.features.feature import CorpusFeature
from somax.runtime.memory_state import MemoryState

""" Keys correspond to parent module names, ex. "pitch" or "chroma". """
EventParameterDict = Dict[str, List[CorpusFeature]]


class Note:
    def __init__(self, pitch: int, velocity: int, channel: int, onset: float, duration: float, track: str,
                 absolute_onset: float, absolute_duration: float):
        self.logger = logging.getLogger(__name__)
        self.pitch: int = pitch
        self.velocity: int = velocity
        self.channel: int = channel
        self.onset: float = onset  # in ticks in relation to CorpusEvent onset
        self.duration: float = duration  # in ticks
        self.track: str = track  # name of MIDI track note originated from
        self.absolute_onset: float = absolute_onset  # in milliseconds in relation to CorpusEvent onset
        self.absolute_duration: float = absolute_duration  # in milliseconds

    @classmethod
    def from_raw(cls, raw_note: pd.Series, parent_event_relative_onset: float, parent_event_absolute_onset: float):
        return cls(pitch=raw_note[Keys.PITCH],
                   velocity=raw_note[Keys.VELOCITY],
                   channel=raw_note[Keys.CHANNEL],
                   onset=raw_note[Keys.REL_ONSET] - parent_event_relative_onset,
                   duration=raw_note[Keys.REL_DURATION],
                   track=raw_note[Keys.TRACK_NAME],
                   absolute_onset=raw_note[Keys.ABS_ONSET] - parent_event_absolute_onset,
                   absolute_duration=raw_note[Keys.ABS_DURATION])

    @classmethod
    def relative_to(cls, note: 'Note', old_parent_onset: float, new_parent_onset: float,
                    old_parent_abs_onset: float, new_parent_abs_onset: float):
        return cls(pitch=note.pitch,
                   velocity=note.velocity,
                   channel=note.channel,
                   onset=note.onset + old_parent_onset - new_parent_onset,
                   duration=note.duration,
                   track=note.track,
                   absolute_onset=note.absolute_onset + old_parent_abs_onset - new_parent_abs_onset,
                   absolute_duration=note.absolute_duration)

    @classmethod
    def from_json(cls, note_dict: Dict[str, Any]):
        return cls(pitch=note_dict["pitch"],
                   velocity=note_dict["velocity"],
                   channel=note_dict["channel"],
                   onset=note_dict["onset"],
                   duration=note_dict["duration"],
                   track=note_dict["track"],
                   absolute_onset=note_dict["absolute_onset"],
                   absolute_duration=note_dict["absolute_duration"])

    def __repr__(self):
        return f"Note(pitch={self.pitch},velocity={self.velocity},channel={self.channel}," \
               f"onset={self.onset},duration={self.duration},..)."

    def __eq__(self, other):
        return self.pitch == other.pitch and self.channel == other.channel

    def encode(self) -> Dict[str, Any]:
        return {"pitch": self.pitch,
                "velocity": self.velocity,
                "channel": self.channel,
                "onset": self.onset,
                "duration": self.duration,
                "track": self.track,
                "absolute_onset": self.absolute_onset,
                "absolute_duration": self.absolute_duration,
                }

    def is_held_from(self, parent_duration: float) -> bool:
        return self.onset + self.duration > parent_duration

    def is_held_to(self) -> bool:
        return self.onset < 0.0


class CorpusEvent:
    def __init__(self, state_index: int, tempo: float, onset: float, absolute_onset: float, bar_number: float,
                 duration: Optional[float] = None, absolute_duration: Optional[float] = None,
                 notes: Optional[List[Note]] = None, features: Optional[Dict[Type[CorpusFeature], CorpusFeature]] = None):
        self.logger = logging.getLogger(__name__)
        self.state_index: int = state_index
        self.tempo: float = tempo
        self.onset: float = onset
        self.absolute_onset: float = absolute_onset
        self.bar_number: float = bar_number

        self.duration: Optional[float] = duration
        self.absolute_duration: Optional[float] = absolute_duration

        self.notes: List[Note] = notes if notes else []
        self.features: Dict[Type[CorpusFeature], CorpusFeature] = features if features else {}
        self.recorded_memory_state: Optional[MemoryState] = None

        # self._labels = {}  # {ClassVar[AbstractLabel]: AbstractLabel}, precompiled for performance

    @classmethod
    def decode(cls, event_dict: Dict[str, Any], classpath_dict: Dict[str, str]) -> 'CorpusEvent':
        """ Raises: KeyError, AttributeError"""
        return CorpusEvent(state_index=event_dict["state_index"],
                           tempo=event_dict["tempo"],
                           onset=event_dict["onset"],
                           absolute_onset=event_dict["absolute_onset"],
                           duration=event_dict["duration"],
                           absolute_duration=event_dict["absolute_duration"],
                           bar_number=event_dict["bar"],
                           notes=[Note.from_json(note_dict) for note_dict in event_dict["notes"]],
                           features=dict([CorpusFeature.from_json(classpath_dict[k], v)
                                          for (k, v) in event_dict["features"].items()])
                           )

    @classmethod
    def incomplete(cls, state_index: int, raw_note: pd.Series) -> 'CorpusEvent':
        """ Create a new CorpusEvent containing a single note where the duration is unknown.
        pd.Series on the format specified in note_matrix.py"""
        event_onset: float = raw_note[Keys.REL_ONSET]
        event_absolute_onset: float = raw_note[Keys.ABS_ONSET]
        bar_number: float = raw_note[Keys.BAR_NUMBER]
        notes: List[Note] = [Note.from_raw(raw_note, event_onset, event_absolute_onset)]
        return cls(state_index, raw_note[Keys.TEMPO], event_onset, event_absolute_onset, bar_number, notes=notes)

    def set_duration(self, end: float, absolute_end: float):
        """ Completes a CorpusEvent constructed with the `incomplete` constructor. """
        self.duration = end - self.onset
        self.absolute_duration = absolute_end - self.absolute_onset

    def append_raw(self, note: pd.Series) -> None:
        self.notes.append(Note.from_raw(note, self.onset, self.absolute_onset))

    def append_notes(self, notes: List[Note], parent_onset: float, parent_abs_onset: float) -> None:
        self.notes.extend([Note.relative_to(n, parent_onset, self.onset, parent_abs_onset, self.absolute_onset)
                           for n in notes])

    def set_feature(self, feature: CorpusFeature):
        self.features[type(feature)] = feature

    def get_feature(self, feature_type: Type[CorpusFeature]) -> CorpusFeature:
        """Raises KeyError"""
        return self.features[feature_type]

    def held_to(self) -> [Note]:
        return [note for note in self.notes if note.is_held_to()]

    def held_from(self) -> [Note]:
        return [note for note in self.notes if note.is_held_from(self.duration)]

    def encode(self, features_dict: Dict[Type['CorpusFeature'], str]) -> Dict[str, Any]:
        return {"state_index": self.state_index,
                "tempo": self.tempo,
                "onset": self.onset,
                "absolute_onset": self.absolute_onset,
                "bar": self.bar_number,
                "duration": self.duration,
                "absolute_duration": self.absolute_duration,
                "notes": [note.encode() for note in self.notes],
                "features": {features_dict[cls]: obj for (cls, obj) in self.features.items()}
                }
        # : Dict[Type[AbstractTrait], AbstractTrait] = event_parameters if event_parameters else {}



from abc import ABC

from somax.runtime.corpus_event import CorpusEvent


class PeakEvent(ABC):
    """ A PeakData is an event matched by a MemorySpace/influence call before it's transformed into a peak"""

    def __init__(self, event: CorpusEvent, transform_hash: int, **_kwargs):
        self.event: CorpusEvent = event
        self.transform_hash: int = transform_hash


class ClassicPeakEvent(PeakEvent):

    def __init__(self, event: CorpusEvent, transform_hash: int, **_kwargs):
        super(ClassicPeakEvent, self).__init__(event, transform_hash)



import logging
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

from somax.scheduler.base_scheduler import BaseScheduler


@dataclass
class Time:
    tick: float
    tempo: float


class Transport(BaseScheduler, ABC):
    BASE_TEMPO = 120.0

    def __init__(self, tempo: float = BASE_TEMPO, *args, **kwargs):
        super().__init__(tempo=tempo, *args, **kwargs)
        self.logger = logging.getLogger(__name__)

    @abstractmethod
    def update_tick(self, **kwargs) -> Time:
        """ raises: may raise TypeError for some subclasses"""
        pass

    @classmethod
    def clone_from(cls, other: 'Transport', *args, **kwargs):
        return cls(tempo=other.tempo, tick=other._tick, *args, **kwargs)

    def set_tempo(self, tempo: float):
        self.tempo = tempo

    def terminate(self):
        self.stop()

    def time(self) -> Time:
        return Time(tick=self._tick, tempo=self.tempo)


class MasterTransport(Transport):
    def __init__(self, tempo: float = Transport.BASE_TEMPO, *args, **kwargs):
        super().__init__(tempo=tempo, *args, **kwargs)
        self._last_callback_time: float = time.time()

    def update_tick(self, **_kwargs) -> Time:
        if self.running:
            self._update_tick()
        return self.time()

    def start(self) -> None:
        self._last_callback_time = time.time()
        self.running = True
        self._update_tick()

    def pause(self) -> None:
        self.running = False

    def stop(self) -> None:
        self.running = False
        self._tick = 0.0

    def _update_tick(self, **_kwargs):
        if self.running:
            t: float = time.time()
            delta_time: float = t - self._last_callback_time
            self._last_callback_time = t
            self._tick += delta_time * self.tempo / 60.0


class SlaveTransport(Transport):
    def __init__(self, tempo: float, *args, **kwargs):
        super().__init__(tempo=tempo, *args, **kwargs)

    def start(self, **kwargs) -> None:
        self.running = True

    def pause(self, **kwargs):
        self.running = False

    def stop(self, **kwargs):
        self.running = False

    def update_tick(self, tick: Optional[float] = None) -> Time:
        """ raises: TypeError if no tick value is provided """
        if tick is None:
            raise TypeError(f"A tick value must be provided for class {self.__class__.__name__}")
        self._tick = tick
        return Time(self.tick, self.tempo)



from abc import ABC
from typing import Dict, Any, Union

import numpy as np

from somax.corpus_builder.chromagram import Chromagram
from somax.features.feature import CorpusFeature, RuntimeFeature


class OnsetChroma(CorpusFeature, RuntimeFeature, ABC):
    def __init__(self, value: Union[np.ndarray, list]):
        super().__init__(value=np.array(value))

    @classmethod
    def decode(cls, trait_dict: Dict[str, Any]) -> 'OnsetChroma':
        return cls(value=np.array(trait_dict["chroma"]))

    def encode(self) -> Dict[str, Any]:
        return {"chroma": self._value.tolist()}

    def value(self) -> np.ndarray:
        return self._value


class BackgroundChroma(OnsetChroma):
    @classmethod
    def analyze(cls, corpus: 'Corpus', **kwargs) -> 'Corpus':
        bg_chromagram: Chromagram = corpus.bg_chromagram
        for event in corpus.events:
            event.set_feature(cls(bg_chromagram.at(event.absolute_onset)))
        return corpus

    # @classmethod
    # def analyze(cls, event: 'CorpusEvent', _fg_spectrogram: Spectrogram, _bg_spectrogram: Spectrogram,
    #             _fg_chromagram: Chromagram, bg_chromagram: Chromagram, **_kwargs):
    #     return cls(bg_chromagram.at(event.absolute_onset))

    @staticmethod
    def keyword() -> str:
        return "chroma"


class ForegroundChroma(OnsetChroma):
    @classmethod
    def analyze(cls, corpus: 'Corpus', **kwargs) -> 'Corpus':
        fg_chromagram: Chromagram = corpus.fg_chromagram
        for event in corpus.events:
            event.set_feature(cls(fg_chromagram.at(event.absolute_onset)))
        return corpus

    # @classmethod
    # def analyze(cls, event: 'CorpusEvent', _fg_spectrogram: Spectrogram, _bg_spectrogram: Spectrogram,
    #             fg_chromagram: Chromagram, _bg_chromagram: Chromagram, **_kwargs):
    #     return cls(fg_chromagram.at(event.absolute_onset))

    @staticmethod
    def keyword() -> str:
        return "foreground_chroma"



import collections
import functools
from abc import ABC
from typing import TypeVar, Union, Dict, Any, Callable, List, Tuple

# TODO: Poor type description
MaxCompatible = TypeVar('MaxCompatible', int, float, bool)
Ranged = Union[MaxCompatible, None]


class HasParameterDict(ABC):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.parameter_dict: Dict[str, Union[Parametric, Parameter, Dict]] = {}


class Parameter(HasParameterDict):

    def __init__(self, default_value: MaxCompatible, min: Ranged, max: Ranged,
                 type_str: str, description: str):
        super().__init__()
        self.value: MaxCompatible = default_value
        self.scope: Tuple[Ranged, Ranged] = (min, max)
        self.type_str: str = type_str
        self.description: str = description

    def max_representation(self) -> Dict:
        # TODO: Remove value from this
        return vars(self)

    def set_value(self, value):
        # TODO: Check range
        self.value = value

    def _parse_parameters(self):
        # Base case: TODO: remove this
        return


class ParamWithSetter(Parameter):
    def __init__(self, default_value: MaxCompatible, min: Ranged, max: Ranged, type_str: str, description,
                 set_function: Callable):
        super().__init__(default_value, min, max, type_str, description)
        self.set_value = set_function


class Parametric(HasParameterDict):

    def __init__(self, **kwargs):
        """ Parameter dict is a dict of dicts (of dicts of ...). Note: only dicts (no lists).
            It should be updated using parameter_dict() whenever parameter info is changed
            (for example, upon creating a streamview, adding a mergeaction or deleting an atom) """
        super(Parametric, self).__init__(**kwargs)
        self.parameter_dict: Dict[str, Union[Parametric, Parameter]] = {}

    def max_representation(self) -> Dict:
        d = {}
        for name, param in self.parameter_dict.items():
            d[name] = param.max_representation()
        return d

    def set_param(self, path: List[str], value: Any):
        """ raises IndexError: if path spec is invalid, for example empty list,
                   ParameterError: if path spec is invalid or if trying to set an object that is not a Parameter.
        """
        param: Parameter = self.get_param(path)
        param.set_value(value)

    def get_param(self, param_path: List[str]) -> Parameter:
        """ raises KeyError """
        param_name: str = param_path.pop(-1)
        parent_dict: Dict[str, Union[Parametric, Parameter]] = functools.reduce(lambda d, key: d[key].parameter_dict,
                                                                                param_path, self.parameter_dict)
        return parent_dict[param_name]

    def _parse_parameters(self) -> {str: Parameter}:
        self.parameter_dict = {}
        param_dict: Dict[str, Union[Parameter, Parametric]] = {}
        for name, variable in vars(self).items():
            # Parse all Parameter and Parametric into dict
            if isinstance(variable, Parameter) or isinstance(variable, Parametric):
                variable._parse_parameters()
                param_dict[name] = variable
            # Parse all Parameter and Parametric inside other dicts (for example MergeAction)
            if isinstance(variable, collections.abc.Mapping):
                for parent, item in variable.items():
                    if isinstance(item, Parameter) or isinstance(item, Parametric):
                        item._parse_parameters()
                        if isinstance(parent, str):
                            param_dict[parent] = item
                        else:
                            param_dict[parent.__name__] = item
        self.parameter_dict = param_dict



from .pitch_features import TopNote, AbstractIntegerPitch, RuntimeIntegerPitch, VirtualFundamental, BassNote
from .chroma_features import ForegroundChroma, BackgroundChroma, OnsetChroma
from .energy_features import MaxVelocity, VerticalDensity
from .spectral_features import OctaveBands


#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

setup(
    author="Joakim Borg",
    author_email='joakim.borg@ircam.fr',
    python_requires='>=3.9',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
    ],
    description="Somax Library",
    # TODO: Package requirements properly and update README (see https://stackoverflow.com/a/33685899)
    install_requires=["numpy", "scipy", "maxosc", "mido", "scikit-learn", "pandas", "python-osc", "matplotlib"],
    license="GNU General Public License v3",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords='somax',
    name='somax',
    packages=find_packages(include=['somax', 'somax.*']),
    url='https://github.com/DYCI2/Somax2',
    version='0.1.1',
    zip_safe=False,
)



from typing import Tuple

import mido
import numpy as np
import pandas as pd

from somax.corpus_builder.matrix_keys import MatrixKeys as Keys
from somax.corpus_builder.midi_parser import MidiParser, BarNumberAnnotation


class NoteMatrix:

    def __init__(self, raw_matrix: pd.DataFrame):
        self.notes: pd.DataFrame = raw_matrix

    def __len__(self) -> int:
        return len(self.notes)

    @classmethod
    def from_midi_file(cls, filepath: str) -> 'NoteMatrix':
        raw_matrix: pd.DataFrame = MidiParser.read_midi(filepath)
        return cls(raw_matrix)

    def to_midi_file(self, title: str, filepath: str, initial_time_signature: Tuple[int, int], ticks_per_beat: int,
                     annotations: BarNumberAnnotation) -> mido.MidiFile:
        return MidiParser.export_midi(self.notes, filepath, title=title, initial_time_signature=initial_time_signature,
                                      ticks_per_beat=ticks_per_beat, annotations=annotations)

    @property
    def pitches(self) -> np.ndarray:
        return self.notes[Keys.PITCH].to_numpy(copy=True)

    @property
    def absolute_onsets(self) -> np.ndarray:
        return self.notes[Keys.ABS_ONSET].to_numpy(copy=True)

    @property
    def absolute_durations(self) -> np.ndarray:
        return self.notes[Keys.ABS_DURATION].to_numpy(copy=True)

    def length_ticks(self) -> float:
        return float(np.max(self.notes[Keys.REL_ONSET] + self.notes[Keys.REL_DURATION]))

    def length_ms(self) -> float:
        return float(np.max(self.notes[Keys.ABS_ONSET] + self.notes[Keys.ABS_DURATION]))

    def split_by_channel(self, channels: Tuple[int]) -> 'NoteMatrix':
        mask = self.notes[Keys.CHANNEL].isin(channels)
        return NoteMatrix(self.notes.copy()[mask])



import inspect
import logging
import sys
from abc import ABC, abstractmethod
from types import ModuleType
from typing import Dict, Type, Any, Optional


class Introspective:

    @classmethod
    def _classes(cls, module: Optional[ModuleType] = None, include_abstract: bool = False) -> Dict[str, Type[Any]]:
        """Returns class objects for all non-abstract classes in the inheriting module."""
        if module is None:
            classes: Dict[str, Any] = dict(
                inspect.getmembers(
                    sys.modules[cls.__module__],
                    lambda member: inspect.isclass(member)
                                   and (include_abstract or not inspect.isabstract(member))
                                   and member.__module__ == cls.__module__)
            )

        else:
            classes: Dict[str, Any] = dict(
                inspect.getmembers(module, lambda m: inspect.isclass(m) and not inspect.isabstract(m))
            )

        return {k.lower(): v for (k, v) in classes.items()}


class StringParsed(Introspective, ABC):

    @classmethod
    @abstractmethod
    def default(cls, **kwargs) -> 'StringParsed':
        """ :raises Should explicitly raise ValueError for classes that shouldn't have a default value """

    @classmethod
    @abstractmethod
    def from_string(cls, class_name: str, **kwargs) -> 'StringParsed':
        """ Should call the protected _from_string() method
            :raises Indirectly raises ValueError if `default()` is not defined.
                    TypeError if not all positional arguments for the class' `__init__` are provided as **kwargs
        """

    @classmethod
    def _from_string(cls, class_name: str, module: Optional[ModuleType] = None, **kwargs) -> 'StringParsed':
        """ :raises ValueError if `default()` is not defined
                    TypeError if not all positional arguments for the class' `__init__` are provided as **kwargs
        """
        if not class_name:
            return cls.default(**kwargs)

        classes: Dict[str, Type[Any]] = cls._classes(module)
        try:
            return classes[class_name.lower()](**kwargs)
        except KeyError:
            logging.getLogger(__name__).warning(f"No class named '{class_name} exists for the "
                                                f"{cls.__module__} module. Using default value.")
            return cls.default(**kwargs)



import asyncio
import logging
import logging.config
import multiprocessing
import os
from importlib import resources
from typing import Any, Optional, List, Tuple

import mido

from somax import settings, log
from somax.classification.classifier import AbstractClassifier
from somax.corpus_builder.corpus_builder import CorpusBuilder
from somax.corpus_builder.midi_parser import BarNumberAnnotation
from somax.corpus_builder.note_matrix import NoteMatrix
from somax.runtime.activity_pattern import AbstractActivityPattern
from somax.runtime.asyncio_osc_object import AsyncioOscObject
from somax.runtime.atom import Atom
from somax.runtime.corpus import Corpus
from somax.runtime.exceptions import DuplicateKeyError, ParameterError, \
    InvalidCorpus, InvalidLabelInput, TransformError
from somax.runtime.influence import FeatureInfluence
from somax.runtime.memory_spaces import AbstractMemorySpace
from somax.runtime.merge_actions import AbstractMergeAction
from somax.runtime.osc_log_forwarder import OscLogForwarder
from somax.runtime.peak_selector import AbstractPeakSelector
from somax.runtime.player import Player
from somax.runtime.process_messages import ControlMessage, TimeMessage, TempoMasterMessage, PlayControl, TempoMessage
from somax.runtime.scale_actions import AbstractScaleAction
from somax.runtime.streamview import Streamview
from somax.runtime.target import Target, SendProtocol
from somax.runtime.transforms import AbstractTransform
from somax.scheduler.agent_scheduler import AgentScheduler
from somax.scheduler.scheduled_event import ScheduledEvent, TempoEvent, MidiEvent, NewStateEvent
from somax.scheduler.scheduled_object import TriggerMode
# TODO: Complete separation Agent/OscAgent where Agent can be initialized and used from the command line
from somax.scheduler.transport import Time


class Agent(multiprocessing.Process):
    def __init__(self, player: Player, recv_queue: multiprocessing.Queue, tempo_send_queue: multiprocessing.Queue,
                 scheduler_tick: float, scheduler_tempo: float, scheduler_running: bool,
                 corpus: Optional[Corpus] = None, **kwargs):
        super().__init__()
        self.logger = logging.getLogger(__name__)
        self.player: Player = player
        self.recv_queue: multiprocessing.Queue = recv_queue
        self.tempo_send_queue: multiprocessing.Queue = tempo_send_queue
        self.scheduler: AgentScheduler = AgentScheduler(player=self.player, tempo=scheduler_tempo,
                                                        initial_tick=scheduler_tick, running=scheduler_running)
        self._enabled: bool = True
        self._terminated: bool = False

        if player.trigger_mode == TriggerMode.AUTOMATIC:
            self.scheduler.add_trigger_event()
        if corpus:  # handle corpus object if passed
            self.player.read_corpus(corpus)

    def terminate(self):
        self._terminated = True


class OscAgent(Agent, AsyncioOscObject):

    def __init__(self, player: Player, recv_queue: multiprocessing.Queue, tempo_send_queue: multiprocessing.Queue,
                 scheduler_tick: float, scheduler_tempo: float, scheduler_running: bool, ip: str, recv_port: int,
                 send_port: int, address: str, corpus_filepath: Optional[str] = None, **kwargs):
        Agent.__init__(self, player=player, recv_queue=recv_queue, tempo_send_queue=tempo_send_queue,
                       scheduler_tick=scheduler_tick, scheduler_tempo=scheduler_tempo,
                       scheduler_running=scheduler_running, **kwargs)
        AsyncioOscObject.__init__(self, recv_port=recv_port, send_port=send_port, ip=ip, address=address, **kwargs)
        self.logger = logging.getLogger(__name__)
        if corpus_filepath:  # handle corpus filepath if passed
            self.read_corpus(corpus_filepath)

    ######################################################
    # ASYNCIO & MAIN LOOP(S)
    ######################################################

    def run(self):
        """ raises: OSError if OSC address already is taken """
        try:
            # TODO: must be handled in every thread. Handle this properly without logging.ini at some point
            with resources.path(log, 'logging.ini') as path:
                logging.config.fileConfig(path.absolute())
            self.logger.addHandler(OscLogForwarder(self.target))
            asyncio.run(self._run())
        except OSError as e:
            self.logger.critical(f"{str(e)}. Could not initialize the agent properly - "
                                 f"please delete it and try with a different OSC receive port.")
            self.terminate()
        except KeyboardInterrupt:
            self.logger.critical(f"Terminating agent '{self.player.name}' due to keyboard interrupt (SIGINT)")
            self.terminate()

    async def _main_loop(self):
        while not self._terminated:
            time: Optional[Time] = None
            while not self.recv_queue.empty():
                message: ControlMessage = self.recv_queue.get()
                if isinstance(message, TimeMessage):
                    time = message.time  # Only process last time message in queue
                if isinstance(message, TempoMasterMessage):
                    self.set_tempo_master(is_master=message.is_master)
                if isinstance(message, ControlMessage):
                    self._process_control_message(message.msg)
            if time is not None:
                self._callback(tick=time.tick, tempo=time.tempo)
            await asyncio.sleep(self.DEFAULT_CALLBACK_INTERVAL)

    def _callback(self, tick: float, tempo: float):
        if self._enabled:
            events: List[ScheduledEvent] = self.scheduler.update_tick(tick=tick, tempo=tempo)
            self._send_events(events)

    def _process_control_message(self, message_type: PlayControl):
        if message_type == PlayControl.START:
            self.start_scheduler()
        if message_type == PlayControl.PAUSE:
            self.pause_scheduler()
        if message_type == PlayControl.STOP:
            self.stop_scheduler()
        if message_type == PlayControl.CLEAR:
            self.clear()
        if message_type == PlayControl.TERMINATE:
            self.terminate()

    def _send_events(self, events: List[ScheduledEvent]):
        for event in events:
            if isinstance(event, TempoEvent):  # Only added to scheduler if tempo master
                self.tempo_send_queue.put(TempoMessage(tempo=event.tempo))
            if isinstance(event, MidiEvent):
                self.target.send(SendProtocol.SEND_MIDI_EVENT, [event.note, event.velocity, event.channel])
            if isinstance(event, NewStateEvent):
                self.target.send(SendProtocol.SEND_STATE_EVENT, [event.corpus_event.state_index,
                                                                 event.applied_transform.renderer_info()])
                notes: List[Tuple[int, int, int]] = [(n.pitch, n.velocity, n.channel) for n in event.corpus_event.notes]
                # " ".join([f"{n[0]} {n[1]} {n[2]}" for n in notes])    # TODO: Remove if flatten works correctly
                self.target.send(SendProtocol.SEND_STATE_ONSET, notes)

    ######################################################
    # SCHEDULER & PLAYBACK CONTROL
    ######################################################

    def start_scheduler(self):
        self.scheduler.start()

    def pause_scheduler(self):
        self.scheduler.pause()

    def stop_scheduler(self):
        self.scheduler.stop()
        self.clear()

    def terminate(self):
        super().terminate()
        self.flush()

    def enabled(self, is_enabled):
        if is_enabled:
            self.logger.debug(f"Agent '{self.player.name}' enabled")
        else:
            self.logger.debug(f"Agent '{self.player.name}' disabled")
            self.flush()
        self._enabled = is_enabled

    def clear(self):
        self.flush()
        self.player.clear()
        self.player.clear_memory()

    def flush(self):
        events: List[ScheduledEvent] = self.scheduler.flush()
        self._send_events(events)

    def set_tempo_master(self, is_master: bool):
        self.scheduler.is_tempo_master = is_master

    ######################################################
    # MAIN RUNTIME FUNCTIONS
    ######################################################

    def influence(self, path: str, feature_keyword: str, value: Any, **kwargs):
        self.logger.debug(f"[influence] called for agent '{self.player.name}' with path '{path}', "
                          f"feature keyword '{feature_keyword}', value '{value}' and kwargs {kwargs}")
        if not self.scheduler.running:
            return
        try:
            influence: FeatureInfluence = FeatureInfluence.from_keyword(feature_keyword, value)
        except ValueError as e:
            self.logger.error(f"{str(e)}. No influence was computed.")
            return

        try:
            path_and_name: List[str] = self._parse_streamview_atom_path(path)
            time: float = self.scheduler.tick
            self.player.influence(path_and_name, influence, time, **kwargs)
        except (AssertionError, KeyError, IndexError, InvalidLabelInput) as e:
            self.logger.error(f"{str(e)} Could not influence target.")
            return
        except InvalidCorpus as e:
            self.logger.debug(repr(e))
            return
        self.logger.debug(f"[influence] Influence successfully completed for agent '{self.player.name}' "
                          f"with path '{path}'.")

    def influence_onset(self):
        if not self.scheduler.running:
            return
        if self.player.trigger_mode == TriggerMode.MANUAL:
            self.logger.debug(f"[influence_onset] Influence onset triggered for agent '{self.player.name}'.")
            self.scheduler.add_trigger_event()

    ######################################################
    # CREATION/DELETION OF STREAMVIEW/ATOM
    ######################################################

    def create_streamview(self, path: str, weight: float = Streamview.DEFAULT_WEIGHT,
                          merge_action: str = "", override: bool = False):
        try:
            path_and_name: List[str] = self._parse_streamview_atom_path(path)
            merge_action: AbstractMergeAction = AbstractMergeAction.from_string(merge_action)
            self.player.create_streamview(path=path_and_name, weight=weight, merge_action=merge_action,
                                          override=override)
            self.logger.info(f"Created streamview with path '{path}'.")
        except (AssertionError, ValueError, KeyError, IndexError, DuplicateKeyError) as e:
            self.logger.error(f"{str(e)}. No streamview was created.")

    def delete_streamview(self, path: str):
        try:
            path_and_name: List[str] = self._parse_streamview_atom_path(path)
            self.player.delete_streamview(path_and_name)
            self.logger.info(f"Deleted streamview with path '{path}'.")
        except (AssertionError, KeyError, IndexError) as e:
            self.logger.error(f"{str(e)} No streamview was deleted.")

    def create_atom(self, path: str = "", weight: float = Atom.DEFAULT_WEIGHT, classifier: str = "",
                    activity_pattern: str = "", memory_space: str = "", self_influenced: bool = False,
                    enabled: bool = True, override: bool = False, **kwargs):
        try:
            path_and_name: List[str] = self._parse_streamview_atom_path(path)
            classifier: AbstractClassifier = AbstractClassifier.from_string(classifier, **kwargs)
            activity_pattern: AbstractActivityPattern = AbstractActivityPattern.from_string(activity_pattern)
            memory_space: AbstractMemorySpace = AbstractMemorySpace.from_string(memory_space)
            self.player.create_atom(path=path_and_name, weight=weight, self_influenced=self_influenced,
                                    classifier=classifier, activity_pattern=activity_pattern,
                                    memory_space=memory_space, enabled=enabled, override=override)
            self.send_atoms()
            # self.logger.info(f"Created atom with path '{path}'.")
        except (AssertionError, ValueError, KeyError, IndexError, DuplicateKeyError) as e:
            self.logger.error(f"{str(e)} No atom was created.")

    def delete_atom(self, path: str):
        try:
            path_and_name: List[str] = self._parse_streamview_atom_path(path)
            self.player.delete_atom(path_and_name)
            self.logger.info(f"Deleted atom with path '{path}'.")
        except (AssertionError, KeyError, IndexError) as e:
            self.logger.error(f"{str(e)} No atom was deleted.")

    ######################################################
    # MODIFY PLAYER/STREAMVIEW/ATOM STATE
    ######################################################

    def set_peak_selector(self, peak_selector: str, verbose: bool = True, **kwargs):
        try:
            peak_selector: AbstractPeakSelector = AbstractPeakSelector.from_string(peak_selector, **kwargs)
            self.player.set_peak_selector(peak_selector)
            if verbose:
                self.logger.info(f"[set_peak_selector] Peak selector set to {type(peak_selector).__name__} "
                                 f"for player '{self.player.name}.")
        except (ValueError, KeyError) as e:
            self.logger.error(f"{str(e)} No peak selector was set.")

    def set_classifier(self, path: str, classifier: str, **kwargs):
        try:
            path_and_name: List[str] = self._parse_streamview_atom_path(path)
            classifier: AbstractClassifier = AbstractClassifier.from_string(classifier, **kwargs)
            self.player.set_classifier(path_and_name, classifier)
            self.logger.debug(f"[set_peak_classifier] Classifier set to {type(classifier).__name__} "
                              f"for player '{self.player.name}' (path='{path}').")
        except (AssertionError, KeyError, ValueError, InvalidCorpus) as e:
            self.logger.error(f"{str(e)} No classifier was set.")

    def set_activity_pattern(self, path: str, activity_pattern: str, **kwargs):
        try:
            path_and_name: List[str] = self._parse_streamview_atom_path(path)
            activity_pattern: AbstractActivityPattern = AbstractActivityPattern.from_string(activity_pattern, **kwargs)
            self.player.set_activity_pattern(path_and_name, activity_pattern)
            self.logger.debug(f"[set_acitivity_pattern] Activity pattern set to {type(activity_pattern).__name__} "
                              f"for player '{self.player.name}.")
        except (AssertionError, KeyError, ValueError) as e:
            self.logger.error(f"{str(e)} No activity pattern was set.")

    def add_transform(self, transform: str, **kwargs):
        try:
            transform: AbstractTransform = AbstractTransform.from_string(transform, **kwargs)
            self.player.add_transform(transform)
            self.logger.debug(f"[add_transform] Added transform {transform} for player '{self.player.name}'.")
        except TransformError as e:
            self.logger.debug(f"{str(e)}. No transform was added.")
        except TypeError as e:
            self.logger.error(f"{str(e)}. Please provide this argument on the form 'argname= value'."
                              f" No transform was added.")

    def remove_transform(self, transform: str, **kwargs):
        try:
            transform: AbstractTransform = AbstractTransform.from_string(transform, **kwargs)
            self.player.remove_transform(transform)
            self.logger.debug(f"Successfully removed transform {transform} from player '{self.player.name}'.")
        except IndexError as e:
            self.logger.error(f"{str(e)} No transform was removed.")
        except TransformError as e:
            self.logger.debug(f"{str(e)}. No transform was removed.")
        except TypeError as e:
            self.logger.error(f"{str(e)}. Please provide this argument on the form 'argname= value'. "
                              f"No transform was removed.")

    def add_scale_action(self, scale_action: str, override: bool = False, verbose: bool = True, **kwargs):
        try:
            scale_action: AbstractScaleAction = AbstractScaleAction.from_string(scale_action, **kwargs)
            self.player.add_scale_action(scale_action, override)
            if verbose:
                self.logger.info(f"Added scale action {repr(scale_action)}")
        except ValueError as e:
            self.logger.error(f"{str(e)}. No scale action was added.")
        except DuplicateKeyError as e:
            self.logger.error(f"{str(e)}. No scale action was added.")

    def remove_scale_action(self, scale_action: str, verbose: bool = True, **kwargs):
        try:
            scale_action: AbstractScaleAction = AbstractScaleAction.from_string(scale_action, **kwargs)
            self.player.remove_scale_action(type(scale_action))
            if verbose:
                self.logger.info(f"Removed scale action {repr(scale_action)}")
        except KeyError as e:
            if verbose:
                self.logger.error(f"Could not remove scale action: {repr(e)}.")

    def read_corpus(self, filepath: str, volatile: bool = False):
        self.logger.info(f"Reading corpus at '{filepath}' for player '{self.player.name}'...")
        if not os.path.exists(filepath):
            self.logger.error(f"The file '{filepath}' does not exist. No corpus was read.")
            return

        try:
            _, file_extension = os.path.splitext(filepath)
            corpus: Corpus = Corpus.from_json(filepath, volatile)

            # TODO[MULTIP]: Handle scheduler accordingly when setting corpus
            self.clear()
            # restart_server: bool = False
            # if self.scheduler.running:
            #     self.scheduler.flush_held()
            #     self.scheduler.pause()
            #     self.clear()
            #     restart_server = True

            self.player.read_corpus(corpus)
            self.logger.info(f"Corpus '{corpus.name}' successfully loaded in player '{self.player.name}'.")
            self.send_current_corpus_info()

            # if restart_server:
            #     self.scheduler.start()

        except (IOError, InvalidCorpus) as e:  # TODO: Missing all exceptions from CorpusBuilder.build()
            self.logger.error(f"{str(e)}. No corpus was read.")

    def set_param(self, path: str, value: Any):
        self.logger.debug(f"[set_param] Attempting to set parameter for player '{self.player.name}' at '{path}' "
                          f"to {value} (type={type(value)})...")
        try:
            path_and_name: List[str] = self._parse_streamview_atom_path(path)
            self.player.set_param(path_and_name, value)
        except (AssertionError, IndexError, ParameterError) as e:
            self.logger.error(f"{str(e)} Could not set parameter.")
        except KeyError as e:
            self.logger.error(f"Could not find {str(e)}. No parameter was set.")

    ######################################################
    # SCHEDULING STATE-RELATED PARAMETERS
    ######################################################

    # TODO: Remove and change into generic set scheduling param
    # TODO: Should also generalize Scheduler.add_trigger_event or some other aspect so that the last lines
    #       of this function are handled by scheduler, not at parsing time

    def set_trigger_mode(self, mode: str):
        try:
            trigger_mode: TriggerMode = TriggerMode.from_string(mode)
        except ValueError as e:
            self.logger.error(f"{repr(e)}. No trigger mode was set.")
            return
        previous_trigger_mode: TriggerMode = self.player.trigger_mode
        self.player.trigger_mode = trigger_mode
        self.flush()

        if trigger_mode == TriggerMode.AUTOMATIC and previous_trigger_mode != trigger_mode:
            self.scheduler.add_trigger_event()
        self.logger.debug(f"[trigger_mode]: Trigger mode set to '{trigger_mode}' for player '{self.player.name}'.")

    def set_held_notes_mode(self, enable: bool):
        self.player.hold_notes_artificially = enable
        self.flush()
        self.logger.debug(f"Held notes mode set to {enable} for player '{self.player.name}'.")

    def set_onset_mode(self, enable: bool):
        self.player.simultaneous_onsets = enable
        self.flush()
        self.logger.debug(f"Simultaneous onset mode set to {enable} for player '{self.player.name}'.")

    ######################################################
    # PRIVATE
    ######################################################

    @staticmethod
    def _parse_streamview_atom_path(path: str) -> List[str]:
        """ :raises AssertionError if `path` is non-empty and not of type `str`. """
        assert isinstance(path, str), "Argument 'path' should be a string."
        if not path:
            return []
        elif settings.PATH_SEPARATOR in path:
            return [s for s in path.split(settings.PATH_SEPARATOR) if s]
        else:
            return [path]

    ######################################################
    # MAX GETTERS
    ######################################################

    # TODO: can be single function with send_corpora
    def get_corpus_files(self, filepath: str):
        # filepath: str = os.path.join(os.path.dirname(__file__), settings.CORPUS_FOLDER)
        corpora: List[Tuple[str, str]] = []
        for file in os.listdir(filepath):
            if any([file.endswith(extension) for extension in CorpusBuilder.CORPUS_FILE_EXTENSIONS]):
                corpus_name, _ = os.path.splitext(file)  # TODO: Not the corpus name that's specified in the json
                corpora.append((corpus_name, os.path.join(filepath, file)))
        self.send_corpora(corpora)

    # TODO: can be single function with send_peaks
    def get_peaks(self):
        self.send_peaks()

    def get_param(self, path: str):
        try:
            parameter_path: List[str] = self._parse_streamview_atom_path(path)
            self.target.send(SendProtocol.PLAYER_SINGLE_PARAMETER, [path, self.player.get_param(parameter_path).value])
        except (IndexError, KeyError):
            self.logger.error(f"Could not get parameter at given path.")
        except (ParameterError, AssertionError) as e:
            self.logger.error(str(e))

    def send_peaks(self):
        peaks_dict, num_recorded_events = self.player.get_peaks()
        for name, count in peaks_dict.items():
            self.target.send(SendProtocol.PLAYER_NUM_PEAKS, [name, count])
        self.target.send(SendProtocol.PLAYER_RECORDED_CORPUS_LENGTH, num_recorded_events)

    def send_corpora(self, corpus_names_and_paths: List[Tuple[str, str]]):
        for corpus in corpus_names_and_paths:
            self.target.send(SendProtocol.PLAYER_CORPUS_FILES, corpus)
        self.target.send(SendProtocol.PLAYER_CORPUS_FILES, Target.WRAPPED_BANG)

    def send_atoms(self):
        atom_names: List[str] = [atom.name for atom in self.player.all_atoms()]
        self.target.send(SendProtocol.INSTANTIATED_ATOMS, atom_names)

    def send_current_corpus_info(self):
        corpus: Optional[Corpus] = self.player.corpus
        if corpus is not None:
            self.target.send(SendProtocol.PLAYER_CORPUS, [corpus.name, corpus.content_type.value, corpus.length()])

    ######################################################
    # OTHER
    ######################################################

    def force_jump(self, index: int):
        try:
            self.player.force_jump(int(index))
            self.scheduler.flush()
        except ValueError as e:
            self.logger.info(f"{str(e)}")

    def export_runtime_corpus(self, folder: str, filename: str, corpus_name: Optional[str] = None,
                              initial_time_signature: Tuple[int, int] = (4, 4), ticks_per_beat: int = 480,
                              annotations: str = BarNumberAnnotation.NONE.value, overwrite: bool = False,
                              use_original_tempo: bool = False):

        filepath = os.path.join(folder, filename)
        if os.path.splitext(filepath)[-1] not in CorpusBuilder.MIDI_FILE_EXTENSIONS:
            filepath += ".mid"
        if os.path.exists(filepath) and not overwrite:
            self.logger.error(f"The file '{filepath}' already exists. No corpus was exported. "
                              f"To override, use 'overwrite= True'.")
            return
        if not os.path.isdir(folder):
            self.logger.error(f"The folder '{folder}' does not exist. No corpus was exported.")
            return

        name: str = corpus_name if corpus_name is not None else filename

        try:
            corpus: Corpus = self.player.export_runtime_corpus(corpus_name, use_original_tempo=use_original_tempo)
        except InvalidCorpus as e:
            self.logger.error(f"{str(e)}. No MIDI data was exported.")
            return

        bar_number_annotations: BarNumberAnnotation = BarNumberAnnotation.from_string(annotations)

        note_matrix: NoteMatrix = corpus.to_note_matrix()
        midi_file: mido.MidiFile = note_matrix.to_midi_file(name, filepath, initial_time_signature, ticks_per_beat,
                                                            bar_number_annotations)
        # midi_file.save(filename=filepath)
        self.logger.info(f"The recorded corpus '{name}' was saved to '{filepath}'.")



import colorsys
from typing import Optional, List, Tuple, Dict, Any

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import patches
from matplotlib.axes import Axes

from somax.corpus_builder.matrix_keys import MatrixKeys as Keys
from somax.corpus_builder.midi_parser import MidiParser
from somax.corpus_builder.note_matrix import NoteMatrix


class NoteMatrixUtils:

    @staticmethod
    def matrix_from_text(notes: np.ndarray, velocities: Optional[np.ndarray] = None,
                         channels: Optional[np.ndarray] = None,
                         durations_tick: Optional[np.ndarray] = None, tempo: float = 120.0) -> NoteMatrix:
        if not velocities:
            velocities = 127 * np.ones_like(notes)
        if not channels:
            channels = np.zeros_like(notes)
        if not durations_tick:
            durations_tick = np.ones_like(notes)
        raw_matrix: np.ndarray = MidiParser.monophonic_from_text(notes, velocities, channels, durations_tick, tempo)
        return NoteMatrix(raw_matrix)

    @staticmethod
    def write_csv(note_matrix: NoteMatrix, filepath: str) -> None:
        note_matrix.notes.to_csv(filepath, index=False, encoding='utf-8')

    @staticmethod
    def matrix_legacy_format(note_matrix: NoteMatrix) -> np.ndarray:
        legacy_matrix: np.ndarray = note_matrix.notes.to_numpy(copy=True)
        legacy_matrix = legacy_matrix[:, [3, 5, 2, 0, 1, 4, 6, 7]]
        return legacy_matrix

    @staticmethod
    def plot(note_matrix: NoteMatrix, axes: Optional[Tuple[Axes, Axes]] = None, x_pixels: int = 100_000,
             saturation: int = 0.7, vibrance: int = 0.8, num_time_indicators: int = 8,
             slice_onset_ticks: Optional[List[float]] = None, top_note: Optional[int] = None,
             bottom_note: Optional[int] = None) -> None:
        # TODO: Implement with proper indexing
        top_note: int = top_note if top_note is not None else int(np.max(note_matrix.notes[Keys.PITCH]))
        bottom_note: int = bottom_note if bottom_note is not None else int(np.min(note_matrix.notes[Keys.PITCH]))
        print(bottom_note)
        duration_ticks: float = note_matrix.length_ticks()

        step_x: float = x_pixels / duration_ticks

        channels: List[int] = sorted([int(ch) for ch in note_matrix.notes[Keys.CHANNEL].unique()])
        channel_colors: Dict[int, Tuple[Any, ...]] = {}
        for i, channel in enumerate(channels):
            channel_colors[channel] = tuple(c for c in colorsys.hsv_to_rgb(i / len(channels), saturation, vibrance))

        image = NoteMatrixUtils._midi_image(note_matrix, x_pixels, top_note, step_x, channel_colors)

        if not axes:
            # type: Tuple[Axes, Axes]
            _, (tempo_ax, notes_ax) = plt.subplots(2, 1, gridspec_kw={'height_ratios': [1, 5]})
        else:
            tempo_ax, notes_ax = axes
        # fig.tight_layout(rect=[0.2, 0.03, 1, 0.95])
        # fig: Figure = plt.figure(constrained_layout=False, tight_layout=False)
        # gridspec = fig.add_gridspec(6, 1, wspace=0.05)
        # notes_ax: AxesImage = fig.add_subplot(gridspec[:-1, :])
        # TODO: Fix title position
        notes_ax.imshow(image, aspect='auto', origin='lower', extent=[0, duration_ticks, 0, top_note])
        notes_ax.set_ylim(bottom_note, top_note)
        notes_ax.set_ylabel("Note Number")
        notes_ax.set_xlabel("Time [ticks]")
        # TODO: Incorrect handling. relationship between ticks and tempo is not just linear scaling - varies with tempo
        # upper_ax: Axes = notes_ax.twiny()
        # upper_ax.set_xbound(0, duration_s)
        # upper_ax.set_xlabel("Time [ms]")
        legends = [patches.Patch(color=col, label=f"Ch. {int(ch)}") for (ch, col) in channel_colors.items()]
        notes_ax.legend(handles=legends)

        if slice_onset_ticks is not None:
            for slice_tick in slice_onset_ticks:
                notes_ax.axvline(slice_tick, dashes=[1, 1], linewidth=1, c='k', alpha=0.2)

        # tempo_ax: Axes = fig.add_subplot(gridspec[-1, :])
        tempo_ax.plot(note_matrix.notes[Keys.REL_ONSET], note_matrix.notes[Keys.TEMPO])
        tempo_ax.set_ylabel("Tempo\n[bpm]", fontsize=8)
        tempo_ax.set_xlim(left=0, right=duration_ticks)
        tempo_ax.set_ylim(bottom=0, top=int(np.max(note_matrix.notes[Keys.TEMPO] * 1.2)))
        tempo_ax.tick_params(labelsize=8)

        duration_ms: float = note_matrix.length_ms()
        ms_thresholds: np.ndarray = np.arange(num_time_indicators - 2) / (num_time_indicators - 2) * duration_ms
        for threshold in ms_thresholds:
            idx: int = int(np.argmax((note_matrix.notes[Keys.ABS_ONSET].to_numpy() >= threshold)))
            x_pos = note_matrix.notes[Keys.REL_ONSET][idx]
            m, s = divmod(note_matrix.notes[Keys.ABS_ONSET][idx] / 1000.0, 60)
            h, m = divmod(m, 60)
            label: str = (f"{int(h)}h" if h else "") + (f"{int(m):02d}'" if m else "") + f"{int(s):02d}\""
            notes_ax.axvline(x_pos, ymin=0.98, linewidth=1, c='k')
            # notes_ax.axvline(x_pos, dashes=[1, 1], linewidth=1, c='k', alpha=0.2)
            notes_ax.annotate(label, xy=(x_pos, top_note + 1), va="center", ha="center", annotation_clip=False,
                              fontsize=6)
        # TODO: Clean duplication
        x_pos = duration_ticks
        m, s = divmod(duration_ms / 1000.0, 60)
        h, m = divmod(m, 60)
        label = (f"{int(h)}h" if h else "") + (f"{int(m):02d}'" if m else "") + f"{int(s):02d}\""
        notes_ax.annotate(label, xy=(x_pos, top_note + 1), va="center", ha="center", annotation_clip=False, fontsize=6)

        plt.show()

    @staticmethod
    def _midi_image(note_matrix: NoteMatrix, num_x_px: int, num_y_px: int, step_x: float,
                    color_dict: Dict[int, Tuple[Any, ...]]) -> np.ndarray:
        # shape: (N, 2). Each row is a pair of (start, end) tick.
        x_coords: np.ndarray = np.asarray(
            np.column_stack((note_matrix.notes[Keys.REL_ONSET],
                             note_matrix.notes[Keys.REL_ONSET] + note_matrix.notes[Keys.REL_DURATION]))
            * step_x, dtype=np.int32)

        # shapes: (N,)
        y_coords: np.ndarray = np.asarray(note_matrix.notes[Keys.PITCH], dtype=np.int32)
        colors: np.ndarray = np.asarray(note_matrix.notes[Keys.CHANNEL], dtype=np.float32)
        alphas: np.ndarray = np.asarray(note_matrix.notes[Keys.VELOCITY], dtype=np.float32) / 127

        image: np.ndarray = np.ones((num_y_px + 1, num_x_px + 1, 4))

        for x, y, c, a in zip(x_coords, y_coords, colors, alphas):
            r, g, b = color_dict[c]
            image[y, x[0]:x[1], 0] *= r
            image[y, x[0]:x[1], 1] *= g
            image[y, x[0]:x[1], 2] *= b
            image[y, x[0]:x[1], 3] *= a

        return image



import copy
import logging
from abc import ABC, abstractmethod
from typing import List, Type, Union, Any

import numpy as np

from somax.features.chroma_features import OnsetChroma
from somax.features.feature import AbstractFeature
from somax.features.pitch_features import AbstractIntegerPitch
from somax.runtime.corpus_event import CorpusEvent
from somax.runtime.exceptions import TransformError, TransformIdentityError
from somax.utils.introspective import StringParsed


class AbstractTransform(StringParsed, ABC):
    def __init__(self):
        self.logger = logging.getLogger(__name__)

    @abstractmethod
    def __eq__(self, other):
        """ Used by TransformHandler to determine whether a transform already exists """

    @staticmethod
    @abstractmethod
    def valid_features() -> List[Type[AbstractFeature]]:
        """"""

    @abstractmethod
    def renderer_info(self) -> Any:
        """ Value sent to renderer describing how to apply the transform to the event if necessary """

    @abstractmethod
    def apply(self, obj: Union[CorpusEvent, AbstractFeature], **kwargs) -> Union[CorpusEvent, AbstractFeature]:
        """ Note: Should return a deepcopy of the event """

    @abstractmethod
    def inverse(self, obj: Union[CorpusEvent, AbstractFeature], **kwargs) -> Union[CorpusEvent, AbstractFeature]:
        """"""

    @classmethod
    def default(cls, **kwargs) -> 'AbstractTransform':
        return NoTransform()

    @classmethod
    def from_string(cls, transform: str, **kwargs) -> 'AbstractTransform':
        """ :raises TypeError if not all positional arguments for the transform's `__init__` are provided as **kwargs"""
        try:
            return cls._from_string(transform, **kwargs)
        except TransformIdentityError:
            # Parameters given duplicates the NoTransform class
            return cls.default()


class NoTransform(AbstractTransform):
    def __init__(self):
        super().__init__()

    def __eq__(self, other):
        return type(self) == type(other)

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    @staticmethod
    def valid_features() -> List[Type[AbstractFeature]]:
        return AbstractFeature.classes(include_abstract=True)

    def renderer_info(self) -> Any:
        return None

    def apply(self, obj: Union[CorpusEvent, AbstractFeature], **kwargs) -> Union[CorpusEvent, AbstractFeature]:
        return obj

    def inverse(self, obj: Union[CorpusEvent, AbstractFeature], **kwargs) -> Union[CorpusEvent, AbstractFeature]:
        return obj


class RedundantTransform(AbstractTransform):
    """ Class for representing transforms removed by the user in the interface while peaks using the transform
        may still exist in the runtime state. """

    def __init__(self, contained_transform: AbstractTransform):
        super().__init__()
        self._contained_transform: AbstractTransform = contained_transform

    def __eq__(self, other):
        return False

    @staticmethod
    def valid_features() -> List[Type[AbstractFeature]]:
        return []

    def renderer_info(self) -> Any:
        return None

    def apply(self, obj: Union[CorpusEvent, AbstractFeature], **kwargs) -> Union[CorpusEvent, AbstractFeature]:
        return self._contained_transform.apply(obj, **kwargs)

    def inverse(self, obj: Union[CorpusEvent, AbstractFeature], **kwargs) -> Union[CorpusEvent, AbstractFeature]:
        return self._contained_transform.inverse(obj, **kwargs)


class TransposeTransform(AbstractTransform):
    def __init__(self, semitones: int):
        super().__init__()
        if semitones == 0:
            # if instantiated through `from_string`: this error is caught and returns a NoTransform object
            raise TransformIdentityError("A transposition of 0 semitones is equivalent to the NoTransform class")
        self.semitones = semitones

    def __eq__(self, other):
        return type(self) == type(other) and self.semitones == other.semitones

    def __repr__(self):
        return f"{self.__class__.__name__}(semitones={self.semitones})"

    def renderer_info(self) -> Any:
        return self.semitones

    @staticmethod
    def valid_features() -> List[Type[AbstractFeature]]:
        return [AbstractIntegerPitch, OnsetChroma]

    def apply(self, obj: Union[CorpusEvent, AbstractFeature], **kwargs) -> Union[CorpusEvent, AbstractFeature]:
        if isinstance(obj, CorpusEvent):
            event: CorpusEvent = copy.deepcopy(obj)
            for (key, feature) in event.features.items():
                try:
                    event.features[key] = self.apply(feature)
                except TransformError:
                    continue
            for note in event.notes:
                note.pitch += self.semitones
            return event
        elif isinstance(obj, AbstractIntegerPitch):
            pitch: int = obj.value() + self.semitones
            return obj.__class__(value=pitch)
        elif isinstance(obj, OnsetChroma):
            chroma: np.ndarray = np.roll(obj.value(), self.semitones % 12)
            return obj.__class__(value=chroma)
        else:
            raise TransformError(f"Could not apply transform {type(self).__name__} to object {obj}. "
                                 f"Valid feature are {self.valid_features()}")

    def inverse(self, obj: Union[CorpusEvent, AbstractFeature], **kwargs) -> Union[CorpusEvent, AbstractFeature]:
        if isinstance(obj, CorpusEvent):
            event: CorpusEvent = copy.deepcopy(obj)
            for (key, feature) in event.features.items():
                try:
                    event.features[key] = self.inverse(feature)
                except TransformError:
                    continue
            for note in event.notes:
                note.pitch -= self.semitones
        elif isinstance(obj, AbstractIntegerPitch):
            pitch: int = obj.value() - self.semitones
            return obj.__class__(value=pitch)
        elif isinstance(obj, OnsetChroma):
            chroma: np.ndarray = np.roll(obj.value(), -(self.semitones % 12))
            return obj.__class__(value=chroma)
        raise TransformError(f"Could not apply inverse transform {type(self).__name__} to object {obj}. "
                             f"Valid feature are {self.valid_features()}")



from typing import Tuple

from somax.corpus_builder.note_matrix import NoteMatrix
from somax.runtime.corpus import Corpus


class CorpusExporter:
    def __init__(self):
        pass

    def export(self, corpus: Corpus, file_path: str, time_signature: Tuple[int, int] = (4, 4)):
        note_matrix: NoteMatrix = corpus.to_note_matrix()
        note_matrix.to_midi_file()
